//!
//!

use std::str::FromStr;

use super::super::super::compositions;
use super::super::super::repositories;
use beteran_common_rust as bcr;
use beteran_protobuf_rust as bpr;
use diesel::{
  r2d2::{ConnectionManager, Pool},
  PgConnection,
};
use prost::Message;

///
pub struct Service {
  connection_broker: nats::asynk::Connection,
  queue_broker: String,
  pool: Pool<ConnectionManager<PgConnection>>,
  site_composition: compositions::site::composition::Composition,
  member_composition: compositions::member::composition::Composition,
  member_repository: repositories::member::repository::Repository,
  member_session_repository: repositories::member_session::repository::Repository,
}

impl std::fmt::Debug for Service {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("Service of service.member.service.identity")
      .finish()
  }
}

impl Service {
  ///
  pub fn new(
    connection_broker: nats::asynk::Connection,
    queue_broker: String,
    pool: Pool<ConnectionManager<PgConnection>>,
  ) -> Service {
    Service {
      connection_broker,
      queue_broker,
      pool,
      site_composition: compositions::site::composition::Composition::new(),
      member_composition: compositions::member::composition::Composition::new(),
      member_repository: repositories::member::repository::Repository::new(),
      member_session_repository: repositories::member_session::repository::Repository::new(),
    }
  }

  pub async fn subscribe(&self) -> std::result::Result<(), std::boxed::Box<dyn std::error::Error>> {
    futures::try_join!(
      self.list_members(),
      self.get_member(),
      self.get_member_by_username(),
      self.event_after_signin()
    )
    .map(|_| ())
  }

  fn check_site(
    &self,
    conn: &diesel::PgConnection,
    url: Option<String>,
    site_id: uuid::Uuid,
  ) -> Result<repositories::site::models::Site, bcr::error::rpc::Error> {
    match self
      .site_composition
      .select_by_url(conn, url, site_id)
      .map_err(|e| {
        bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
          code: bpr::protobuf::rpc::Error::SERVER_00,
          message: format!("server {}", e),
          data: None,
        })
      })? {
      Some(s) => Ok(s),
      None => Err(bcr::error::rpc::Error::InvalidParams(
        bcr::error::rpc::InvalidParams {
          message: "invalid site_url information".to_string(),
          detail: bcr::error::rpc::InvalidParamsDetail {
            location: "request".to_string(),
            param: "client.site_url".to_string(),
            value: "".to_string(),
            error_type: bcr::error::rpc::InvalidParamsType::None,
            message: "".to_string(),
          },
        },
      )),
    }
  }

  async fn list_members(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member::member::SUBJECT_LIST_MEMBERS,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let req = bpr::ss::member::member::ListMembersRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let client = match req.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        let find_all = repositories::member::models::FindAll {
          site_id: None,
          member_class_id: None,
          member_level_id: None,
          referrer_member_id: None,
          username_like: None,
          nickname_like: None,
          mobile_phone_number_like: None,
          last_signined_ip: None,
          state: None,
          deleted_at: None,
          pagination: req
            .pagination
            .as_ref()
            .map(bcr::models::pagination::Pagination::from),
          sorts: Some(
            req
              .sorts
              .iter()
              .map(beteran_common_rust::models::pagination::Sort::from)
              .collect(),
          ),
        };

        let m = self
          .member_repository
          .select_all_count(&conn, &find_all)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        let member_list = self
          .member_composition
          .select_all(&conn, &find_all)
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        message
          .respond(
            bpr::ss::member::member::ListMembersResponse {
              error: None,
              result: Some(bpr::ss::member::member::list_members_response::Result {
                members: member_list
                  .iter()
                  .map(bpr::models::member::MemberModel::from)
                  .collect(),
              }),
            }
            .encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::ss::member::member::ListMembersResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn get_member(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member::member::SUBJECT_GET_MEMBER,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let req = bpr::ss::member::member::GetMemberRequest::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let id = uuid::Uuid::from_str(req.id.as_str()).map_err(|e| {
          bcr::error::rpc::Error::InvalidParams(bcr::error::rpc::InvalidParams {
            message: "invalid id param".to_string(),
            detail: bcr::error::rpc::InvalidParamsDetail {
              location: "request".to_string(),
              param: "id".to_string(),
              value: req.id.clone(),
              error_type: bcr::error::rpc::InvalidParamsType::Required,
              message: e.to_string(),
            },
          })
        })?;

        let client = match req.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        let m = self.member_composition.select(&conn, id).map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        message
          .respond(
            bpr::ss::member::member::GetMemberResponse {
              error: None,
              result: Some(bpr::ss::member::member::get_member_response::Result {
                member: m.map(|d| bpr::models::member::MemberModel::from(&d)),
              }),
            }
            .encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::ss::member::member::GetMemberResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn get_member_by_username(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member::member::SUBJECT_GET_MEMBER_BY_USERNAME,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let req =
          bpr::ss::member::member::GetMemberByUsernameRequest::decode(message.data.as_slice())
            .map_err(|e| {
              bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
                message: format!("invalid request: {}", e),
              })
            })?;

        let client = match req.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        let m = self
          .member_composition
          .select_by_username(&conn, req.username.as_str())
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        message
          .respond(
            bpr::ss::member::member::GetMemberByUsernameResponse {
              error: None,
              result: Some(
                bpr::ss::member::member::get_member_by_username_response::Result {
                  member: m.map(|d| bpr::models::member::MemberModel::from(&d)),
                },
              ),
            }
            .encode_to_vec(),
          )
          .await
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        message
          .respond(
            bpr::ss::member::member::GetMemberByUsernameResponse {
              error: Some(bpr::protobuf::rpc::Error::from(e)),
              result: None,
            }
            .encode_to_vec(),
          )
          .await?;
      }
    }

    Ok(())
  }

  async fn event_after_signin(&self) -> Result<(), Box<dyn std::error::Error>> {
    let s = self
      .connection_broker
      .queue_subscribe(
        bpr::ss::member::identity::EVENT_SUBJECT_AFTER_SIGNIN,
        self.queue_broker.as_str(),
      )
      .await?;

    while let Some(message) = s.next().await {
      if let Err(e) = async {
        let event = bpr::ss::member::identity::AfterSigninEvent::decode(message.data.as_slice())
          .map_err(|e| {
            bcr::error::rpc::Error::InvalidRequest(bcr::error::rpc::InvalidRequest {
              message: format!("invalid request: {}", e),
            })
          })?;

        let client = match event.client {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let member = match event.member {
          Some(c) => c,
          None => {
            return Err(bcr::error::rpc::Error::InvalidParams(
              bcr::error::rpc::InvalidParams {
                message: "invalid client information".to_string(),
                detail: bcr::error::rpc::InvalidParamsDetail {
                  location: "request".to_string(),
                  param: "client".to_string(),
                  value: "".to_string(),
                  error_type: bcr::error::rpc::InvalidParamsType::Required,
                  message: "".to_string(),
                },
              },
            ));
          }
        };

        let id = uuid::Uuid::from_str(member.id.as_str()).map_err(|e| {
          bcr::error::rpc::Error::InvalidParams(bcr::error::rpc::InvalidParams {
            message: "invalid member.id param".to_string(),
            detail: bcr::error::rpc::InvalidParamsDetail {
              location: "request".to_string(),
              param: "member.id".to_string(),
              value: member.id.clone(),
              error_type: bcr::error::rpc::InvalidParamsType::Required,
              message: e.to_string(),
            },
          })
        })?;

        let conn = self.pool.get().map_err(|e| {
          bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
            code: bpr::protobuf::rpc::Error::SERVER_00,
            message: format!("server {}", e),
            data: None,
          })
        })?;

        let last_signined_at = chrono::Utc::now().timestamp();

        self
          .member_repository
          .update_last_signined_ip(
            &conn,
            id,
            &repositories::member::models::ModifyMember4LastSignined {
              last_signined_ip: client.client_ip,
              last_signined_at,
            },
          )
          .map_err(|e| {
            bcr::error::rpc::Error::Server(bcr::error::rpc::Server {
              code: bpr::protobuf::rpc::Error::SERVER_00,
              message: format!("server {}", e),
              data: None,
            })
          })?;

        Ok::<(), bcr::error::rpc::Error>(())
      }
      .await
      {
        println!("error: {}", e);
      }
    }

    Ok(())
  }
}
